
################################################################################################
################# ------------------- SIMULATION FUNCTIONS ------------------- #################
################################################################################################

# Simulation of a panel of countries
function simul_panel_fun(cm::CoreModule, asm::AshocksModuleOBC, agg_state, capN, dropT)

    srand(5678) # Setting the seed

    YEMsim  = agg_state[:, 1]
    ωsim  = agg_state[:, 2]
    aDMpsim = agg_state[:, 3]
    capT    = size(YEMsim, 1)

    ### Simulate Idiosyncratic Shocks and Variables
    shocks_z  = rand(Normal(0.0, cm.σz), capT-1, capN)
    zsim      = zeros(capT,   capN)
    bsim      = zeros(capT+1, capN)
    ysim      = zeros(capT,   capN)
    def_sim   = zeros(capT+1, capN)
    spsim     = zeros(capT,   capN)
    csim      = zeros(capT, capN)
    bpy_sim   = zeros(capT,   capN)
    retEM     = zeros(capT,   capN)

    max_q = maximum(asm.q_mat)
    min_q = minimum(asm.q_mat)
    min_b = cm.bgrid[1]
    
    for ic=1:capN
        for tt=1:capT-1
            zsim[tt+1, ic] = cm.ρz*zsim[tt, ic] + shocks_z[tt, ic]
        end
    end

    knots   = (cm.bgrid, cm.zgrid, asm.YEMgrid, cm.ωgrid, cm.aDMgrid)
    itp_q   = interpolate(knots, asm.q_mat,  Gridded(Linear()))
    itp_bp  = interpolate(knots, asm.bprime_mat, Gridded(Linear()))
    itp_def = interpolate(knots, asm.defmat,     Gridded(Linear()))

    for ic=1:capN

        qm1_v = 1.0

        for tt=1:capT
            z_v     = zsim[tt, ic]
            b_v     = bsim[tt, ic]
            YEMv    = YEMsim[tt]
            ωv    = ωsim[tt]
            aDMp_v  = aDMpsim[tt] # it is a_{DM,t+1}!            

            if def_sim[tt, ic] == 0.0
                df_prob = min( max(0.0, itp_def[b_v, z_v, YEMv, ωv, aDMp_v]), 1.0)
                rand_v  = rand()
                df_chce = 1*(rand_v<df_prob)

                if df_chce == 0.0
                    y_v      = exp(z_v+YEMv)
                    bp_v     = max(min_b, itp_bp[b_v, z_v, YEMv, ωv, aDMp_v] )
                    q_v      = min( max( min_q, itp_q[bp_v, z_v, YEMv, ωv, aDMp_v] ), max_q)
                    sp_v     = (1.0/q_v+cm.ξ-1.0) - (1.0/cm.βLD-1.0)
                    c_v      = max(y_v - b_v + q_v*(bp_v-cm.ξ*b_v), 1e-14)
                    def_sim[tt+1, ic] = 0.0
                    bpy_sim[tt, ic]   = bp_v*q_v / y_v
                else
                    y_v  = max( exp(z_v+YEMv)*(1.0 - cm.d0*exp(z_v+YEMv)^cm.d1), 1e-14 )
                    bp_v = 0.0
                    q_v  = 0.0
                    sp_v = 0.0
                    c_v  = y_v
                    def_sim[tt, ic]   = 1.0
                    def_sim[tt+1, ic] = 1.0
                    bpy_sim[tt, ic]   = 0.0
                end
            else
                rand_v = rand(Uniform(0.0 ,1.0))
                if rand_v <= cm.θ # out of default today
                    y_v      = exp(z_v+YEMv)
                    bp_v     = max(min_b, itp_bp[b_v, z_v, YEMv, ωv, aDMp_v] )
                    q_v      = min( max( min_q, itp_q[bp_v, z_v, YEMv, ωv, aDMp_v] ), max_q)
                    sp_v     = (1.0/q_v+cm.ξ-1.0) - (1.0/cm.βLD-1.0)
                    c_v      = max(y_v - b_v + q_v*(bp_v-cm.ξ*b_v), 1e-14)
                    def_sim[tt, ic]   = 0.0
                    def_sim[tt+1, ic] = 0.0
                    bpy_sim[tt, ic]   = bp_v*q_v / y_v
                else
                    y_v  = max( exp(z_v+YEMv)*(1.0 - cm.d0*exp(z_v+YEMv)^cm.d1), 1e-14 )
                    bp_v = 0.0
                    q_v  = 0.0
                    sp_v = 0.0
                    c_v  = y_v
                    def_sim[tt+1, ic] = 1.0
                    bpy_sim[tt, ic]   = 0.0
                end
            end

            ysim[tt, ic]   = y_v
            bsim[tt+1, ic] = bp_v
            spsim[tt, ic]  = sp_v
            csim[tt, ic]   = c_v
            retEM[tt, ic]  = (1-def_sim[tt, ic])*(1.0+cm.ξ*q_v)/qm1_v

            qm1_v = q_v
        end
    end

    return(zsim[dropT+1:capT, :], bsim[dropT+1:capT, :], def_sim[dropT+1:capT, :], spsim[dropT+1:capT, :], ysim[dropT+1:capT, :], csim[dropT+1:capT, :], bpy_sim[dropT+1:capT, :], retEM[dropT+1:capT, :])
end




# Simluation of market value of net worth
function simulate_mktval(cm::CoreModule, asm::AshocksModuleOBC, ω_hist, ADM_hist, NW_hist)

    vs_calib = zeros(length(ω_hist))

    itp_vs    = interpolate((cm.ωgrid, cm.aDMgrid), asm.vs, Gridded(Linear()))

    for t=1:length(ω_hist)
        vs_calib[t]         = itp_vs[ω_hist[t], ADM_hist[t+1]]
    end

    mktval  = vs_calib.*NW_hist

    return mktval
end



###################################################################################################
################ ------------------ MOMENT COMPUTATION FUNCTIONS ----------------- ################
###################################################################################################

# Moments using panel data
function compute_moments_panel_calib(cm::CoreModule, asm::AshocksModuleOBC, YEMsim, ωsim, ADMsim, panel_vars, q_DFRSK, q_IR)

    capT = size(YEMsim, 1)
    capN = size(panel_vars[1], 2)

    zsim, bsim, def_sim, spsim, ysim, csim, bpy_sim = panel_vars[1], panel_vars[2], panel_vars[3], panel_vars[4], panel_vars[5], panel_vars[6], panel_vars[7]

    corr_fun(x1, x2) = cov(x1, x2) / ( std(x1)*std(x2) )

    ind_tmp1 = (def_sim.==0.0).*(spsim.<10_000/10_000)



    knots      = (cm.bgrid, cm.zgrid, asm.YEMgrid, cm.ωgrid, cm.aDMgrid)
    itp_qDFRSK = interpolate(knots, q_DFRSK, Gridded(Linear()))
    itp_qIR    = interpolate((asm.YEMgrid, cm.ωgrid, cm.aDMgrid), q_IR, Gridded(Linear()))


    ### CROSS-SECTIONAL MOMENTS

    csec_mean_DY = zeros(capT-1)
    csec_std_yt, csec_std_ct, csec_std_tb  = copy(csec_mean_DY), copy(csec_mean_DY), copy(csec_mean_DY)
    csec_std_DY, csec_mean_SP, csec_std_SP = copy(csec_mean_DY), copy(csec_mean_DY), copy(csec_mean_DY)
    csec_mean_TB, csec_mean_CA  = copy(csec_mean_DY), copy(csec_mean_DY)
    csec_mean_Y   = copy(csec_mean_DY)
    csec_mean_lnY = copy(csec_mean_DY)

    aggreg_sp = copy(csec_mean_DY)

    for tt=1:capT-1
        yy_vals = ysim[tt, :]
        by_vals = vec(bpy_sim[tt, :])
        sp_vals = vec(spsim[tt, :])
        cc_vals = vec(csim[tt, :])
        ind_tmp = vec(ind_tmp1[tt,:])

        yy_vals = yy_vals[ind_tmp]
        by_vals = by_vals[ind_tmp]
        sp_vals = sp_vals[ind_tmp]
        cc_vals = cc_vals[ind_tmp]
        tb_vals = (yy_vals-cc_vals)./yy_vals

        csec_mean_Y[tt]  = mean(yy_vals)
        csec_mean_DY[tt] = mean(by_vals)
        csec_mean_SP[tt] = mean(sp_vals)
        csec_mean_TB[tt] = mean(tb_vals)
        csec_mean_CA[tt]
        csec_std_DY[tt]  = std(by_vals)
        csec_std_SP[tt]  = std(sp_vals)
        aggreg_sp[tt]    = mean(sp_vals)

        csec_mean_lnY[tt] = mean(log.(yy_vals))
    end

    cs_sd_by       = mean(csec_std_DY)*100
    cs_sd_SP       = mean(csec_std_SP)*10_000
    cs_corr_by_y   = corr_fun(csec_mean_DY, csec_mean_Y)
    cs_corr_SP_y   = corr_fun(csec_mean_SP, csec_mean_Y)
    cs_corr_sdby_y = corr_fun(csec_std_DY, csec_mean_Y)
    cs_corr_sdSP_y = corr_fun(csec_std_SP, csec_mean_Y)

    cs_corr_by_lny   = corr_fun(csec_mean_DY, csec_mean_lnY)
    cs_corr_SP_lny   = corr_fun(csec_mean_SP, csec_mean_lnY)
    cs_corr_sdby_lny = corr_fun(csec_std_DY, csec_mean_lnY)
    cs_corr_sdSP_lny = corr_fun(csec_std_SP, csec_mean_lnY)
    cs_corr_TB_lny   = corr_fun(csec_mean_TB, csec_mean_lnY)

    out_cross_section = [cs_sd_by; cs_sd_SP; cs_corr_by_y; cs_corr_SP_y; cs_corr_sdby_y; cs_corr_sdSP_y; cs_corr_by_lny; cs_corr_SP_lny; cs_corr_sdby_lny; cs_corr_sdSP_lny; cs_corr_TB_lny]



    ### TIME SERIES MOMENTS

    corr_sp_Y = zeros(capN)
    def_freq  = zeros(capN)

    # def_sim = def_sim[1:capT-1, :]

    mean_DY, mean_spread, std_spread, std_tb, sdci_sdyi, corr_ci_yi, sd_TBi_yi, corr_TBi_yi, corr_spi_SP = 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0

    mean_SP_DFRSK, mean_SP_TOTAL, mean_SP_IMRSK, std_SP_DFRSK, std_SP_TOTAL, std_SP_IMRSK = 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
    mean_SP_IR, mean_SP_PR, std_SP_IR, std_SP_PR = 0.0, 0.0, 0.0, 0.0
    std_tb2, sd_TBi_yi2, corr_TBi_yi2, corr_spi_lnci, corr_spi_TBi = 0.0, 0.0, 0.0, 0.0, 0.0
    ln_sdci_sdyi, ln_sd_TBi_yi, ln_corr_ci_yi, ln_corr_TBi_yi, ln_corr_SP_yi = 0.0, 0.0, 0.0, 0.0, 0.0

    ind_tmp1    = ind_tmp1[1:capT-1, :]

    for ic=1:capN
        sp_vals  = vec(spsim[1:capT-1, ic])
        yt_vals  = vec(ysim[1:capT-1, ic])
        b_vals   = vec(bsim[1:capT-1, ic])
        by_vals  = vec(bpy_sim[1:capT-1, ic])
        ct_vals  = vec(csim[1:capT-1, ic])
        tb_vals  = (yt_vals-ct_vals)./yt_vals

        ind_tmp  = ind_tmp1[:, ic]

        yt_vals  = yt_vals[ind_tmp]
        by_vals  = by_vals[ind_tmp]
        sp_vals  = sp_vals[ind_tmp]
        ct_vals  = ct_vals[ind_tmp]
        tb_vals  = tb_vals[ind_tmp]
        tb_vals2 = (yt_vals-ct_vals)./yt_vals

        mean_DY     += mean(by_vals)
        mean_spread += mean(sp_vals)
        std_spread  += std(sp_vals)
        std_tb      += std(tb_vals)
        sdci_sdyi   += std(ct_vals) / std(yt_vals)
        sd_TBi_yi   += std(tb_vals) / std(yt_vals)
        corr_ci_yi  += corr_fun(ct_vals, yt_vals)
        corr_TBi_yi += corr_fun(tb_vals, yt_vals)
        corr_spi_SP += corr_fun(sp_vals, aggreg_sp[ind_tmp])

        ln_sdci_sdyi   += std(log.(ct_vals)) / std(log.(yt_vals))
        ln_sd_TBi_yi   += std(tb_vals) / std(log.(yt_vals))
        ln_corr_ci_yi  += corr_fun(log.(ct_vals), log.(yt_vals))
        ln_corr_TBi_yi += corr_fun(tb_vals, log.(yt_vals))
        ln_corr_SP_yi  += corr_fun(sp_vals, log.(yt_vals))

        corr_spi_lnci  += corr_fun(sp_vals, log.(ct_vals))
        corr_spi_TBi   += corr_fun(sp_vals, tb_vals)

        std_tb2       += std(tb_vals2)
        sd_TBi_yi2    += std(tb_vals2) / std(yt_vals)
        corr_TBi_yi2  += corr_fun(tb_vals2, yt_vals)

        corr_sp_Y[ic] = corr_fun(sp_vals, yt_vals)
        
        cum_num = 0.0; cum_den = 0.0

        SPTOT_nodef, SPDFRSK_nodef, SPIR_nodef = [], [], []

        for tt=2:capT-1
            defm1 = def_sim[tt-1, ic]
            deft  = def_sim[tt, ic]

            if (defm1 == 0.0 && deft == 1.0)
                cum_num += 1.0
            end
            if defm1==0.0
                cum_den += 1.0
            end

            if deft==0.0
                bp_v = bsim[tt+1, ic]
                z_v  = zsim[tt, ic]
                qDFRSK_v     = itp_qDFRSK[bp_v, z_v, YEMsim[tt], ωsim[tt], ADMsim[tt+1]]
                qIR_v        = itp_qIR[YEMsim[tt], ωsim[tt], ADMsim[tt+1]]
                spDFRSK_v    = (1.0/qDFRSK_v + cm.ξ).*cm.βLD - 1.0
                spIR_v       = (1.0/qIR_v + cm.ξ).*cm.βLD - 1.0
                if spDFRSK_v < 10_000/10_000
                    push!(SPDFRSK_nodef, spDFRSK_v)
                    push!(SPIR_nodef, spIR_v)
                    push!(SPTOT_nodef, spsim[tt, ic])
                end
            end
        end

        def_freq[ic] = cum_num / cum_den

        SPDFRSK_nodef = convert(Array{Float64}, SPDFRSK_nodef)
        SPTOT_nodef   = convert(Array{Float64}, SPTOT_nodef)
        SPIR_nodef    = convert(Array{Float64}, SPIR_nodef)
        SPIMRSK_nodef = SPTOT_nodef - SPDFRSK_nodef
        SP_PureRisk   = SPIMRSK_nodef - SPIR_nodef

        mean_SP_DFRSK += mean(SPDFRSK_nodef)
        mean_SP_TOTAL += mean(SPTOT_nodef)
        mean_SP_IMRSK += mean(SPIMRSK_nodef)
        mean_SP_IR    += mean(SPIR_nodef)
        mean_SP_PR    += mean(SP_PureRisk)
        std_SP_DFRSK  += std(SPDFRSK_nodef)
        std_SP_TOTAL  += std(SPTOT_nodef)
        std_SP_IMRSK  += std(SPIMRSK_nodef)
        std_SP_IR     += std(SPIR_nodef)
        std_SP_PR     += std(SP_PureRisk)
    end

    # Moments - Country Average
    mean_DY     = mean_DY / capN*100
    mean_spread = mean_spread/capN*10_000
    std_spread  = std_spread/capN*10_000
    std_tb      = std_tb/capN
    DF_mean     = mean(def_freq)*100
    sdci_sdyi   = sdci_sdyi / capN
    corr_ci_yi  = corr_ci_yi / capN
    sd_TBi_yi   = sd_TBi_yi / capN
    corr_TBi_yi = corr_TBi_yi / capN
    corr_spi_SP = corr_spi_SP / capN

    std_tb2      = std_tb2/capN
    sd_TBi_yi2   = sd_TBi_yi2 / capN
    corr_TBi_yi2 = corr_TBi_yi2 / capN    

    ln_sdci_sdyi   = ln_sdci_sdyi  /capN
    ln_sd_TBi_yi   = ln_sd_TBi_yi  /capN
    ln_corr_ci_yi  = ln_corr_ci_yi /capN
    ln_corr_TBi_yi = ln_corr_TBi_yi/capN
    ln_corr_SP_yi  = ln_corr_SP_yi/capN

    corr_spi_lnci = corr_spi_lnci / capN
    corr_spi_TBi  = corr_spi_TBi / capN

    out_time_series = [mean_DY; DF_mean; mean_spread; std_spread; std_tb; mean(corr_sp_Y)*100; sdci_sdyi; corr_ci_yi; sd_TBi_yi; corr_TBi_yi; corr_spi_SP; std_tb2; sd_TBi_yi2; corr_TBi_yi2; ln_sdci_sdyi; ln_sd_TBi_yi; ln_corr_ci_yi; ln_corr_TBi_yi; ln_corr_SP_yi; corr_spi_lnci; corr_spi_TBi]
    out_decomp      = [mean_SP_DFRSK/capN*10_000; mean_SP_TOTAL/capN*10_000; mean_SP_IMRSK/capN*10_000; std_SP_DFRSK/capN*10_000; std_SP_TOTAL/capN*10_000; std_SP_IMRSK/capN*10_000]
    out_dcmp_interm = [mean_SP_IR/capN*10_000; mean_SP_PR/capN*10_000; std_SP_IR/capN*10_000; std_SP_PR/capN*10_000; 0.0; 0.0]

    return(out_cross_section, out_time_series, out_decomp, out_dcmp_interm)

    

end





###################################################################################################
################ ------------------- IMPULSE RESPONSE FUNCTIONS ------------------ ################
###################################################################################################

# Lehman Episode
function compute_Lehman_proxy(cm::CoreModule, asm::AshocksModuleOBC, init_val, omg_path, YEM_path, Λerg, data_in, mms_in_mod, mms_in_dat, itp_HYSP, itp_qDFRSK; type_shock=3, full_dyn=false)

    # Create interpolation/extrapolation object
    knots    = (cm.bgrid, cm.zgrid, asm.YEMgrid, cm.ωgrid, cm.aDMgrid)
    itp_def  = interpolate(knots, asm.defmat,     Gridded(Linear()))
    itp_q    = interpolate(knots, asm.q_mat,  Gridded(Linear()))
    itp_bp   = interpolate(knots, asm.bprime_mat, Gridded(Linear()))
    itp_ReDM = interpolate((cm.ωgrid, cm.aDMgrid), asm.ReDM, Gridded(Linear()))
    itp_Red  = interpolate((cm.ωgrid, cm.aDMgrid), asm.Red, Gridded(Linear()))

    capT  = size(omg_path, 1)


    ### Simulation in absence of shocks
    Btil_path_NS, Bhat_path_NS, yield_path_NS, EMSP_path_NS, HYSP_path_NS, NW_path_NS, Ct_path_NS = zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    TB_path_NS, C_Y_path_NS, SPDF_path_NS, defmeas_path_NS = zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    Bt_path_NS = zeros(capT+1)

    ADM_path_NS    = zeros(capT+1) # time t
    D_path_NS      = zeros(capT+1) # time t
    ADM_path_NS[1] = init_val[1]
    D_path_NS[1]   = init_val[2]

    Λerg_mkt = Λerg[cm.num_nodes+1:end]
    sum_Λ   = sum(Λerg_mkt)

    Bt = 0.0
    for i_s=1:cm.num_nodes
        Bt    += Λerg_mkt[i_s] * cm.snodes[i_s, 1]
    end

    Bt_path_NS[1] = Bt * cm.μEM / sum_Λ

    YEM_path_NS = zeros(YEM_path)
    omg_path_NS = ones(capT_Lehman)

    Λt = copy(Λerg)


    # Bounds for finding ADM
    bounds = [cm.aDMgrid[1]*0.9; asm.ADMmax]

    for t=1:capT
        state_in = [YEM_path_NS[t]; omg_path_NS[t]; ADM_path_NS[t]; D_path_NS[t]]
        out_X    = solve_system_calib(cm, asm, state_in, Λt, itp_def, itp_q, itp_bp, itp_ReDM, itp_Red, bounds, itp_qDFRSK)
#      out_X[1] = [aDMp_v; Btilv; Bhatv; Bt1; NW; ReDM; SPv; yield_v; by_avg; sdby; sdSP; Ct; SPDFRK; Dpv; aDMp_out; TB; C_Y]
#                    1       2      3     4    5    6    7     8         9     10     11  12    13     14     15     16  17

        aDMpv, Btilde, Bhat, Bt1, NW, ReDM, SPv, yieldv = out_X[1][1], out_X[1][2], out_X[1][3], out_X[1][4], out_X[1][5], out_X[1][6], out_X[1][7], out_X[1][8]
        by_avg, sdby, sdSP, Ct, TB, C_Y = out_X[1][9], out_X[1][10], out_X[1][11], out_X[1][12], out_X[1][16], out_X[1][17]
        D_path_NS[t+1] = out_X[1][14]
        
        Λt1 = out_X[2]

        yield_path_NS[t] = yieldv
        NW_path_NS[t]    = NW
        Btil_path_NS[t]  = Btilde
        Bhat_path_NS[t]  = Bhat
        EMSP_path_NS[t]  = SPv
        Ct_path_NS[t]    = Ct
        TB_path_NS[t]    = TB*100
        C_Y_path_NS[t]   = C_Y*100
        HYSP_path_NS[t]  = itp_HYSP[omg_path[t], aDMpv]
        Bt_path_NS[t+1]  = Bt1
        SPDF_path_NS[t]  = out_X[1][13]

        ADM_path_NS[t+1] = aDMpv

        Λt1_mkt = Λt1[cm.num_nodes+1:end]
        defmeas_path_NS[t] = 1.0 - sum(Λt1_mkt)

        Λt = copy(Λt1)

    end

    MV_path_NS = simulate_mktval(cm, asm, omg_path_NS, ADM_path_NS, NW_path_NS)

    CA_path_NS = TB_path_NS - (1-cm.ξ)*Bt_path_NS[1:end-1]./exp.(YEM_path_NS)


    ### Simulation given shocks
    Btil_path, Bhat_path, yield_path, EMSP_path, HYSP_path, NW_path, Ct_path = zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    TB_path, C_Y_path, SPDF_path, defmeas_path = zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    Bt_path = zeros(capT+1)

    ADM_path    = zeros(capT+1) # time t
    D_path      = zeros(capT+1) # time t
    ADM_path[1] = init_val[1]
    D_path[1]   = init_val[2]

    Bt_path[1]  = Bt_path_NS[1]

    Λt = copy(Λerg)

    for t=1:capT

        state_in = [YEM_path[t]; omg_path[t]; ADM_path[t]; D_path[t]]
        out_X = solve_system_calib(cm, asm, state_in, Λt, itp_def, itp_q, itp_bp, itp_ReDM, itp_Red, bounds, itp_qDFRSK)


        aDMpv, Btilde, Bhat, Bt1, NW, ReDM, SPv, yieldv = out_X[1][1], out_X[1][2], out_X[1][3], out_X[1][4], out_X[1][5], out_X[1][6], out_X[1][7], out_X[1][8]
        by_avg, sdby, sdSP, Ct, TB, C_Y = out_X[1][9], out_X[1][10], out_X[1][11], out_X[1][12], out_X[1][16], out_X[1][17]
        SPDF_path[t] = out_X[1][13]
        D_path[t+1]  = out_X[1][14]

        Λt1 = out_X[2]

        yield_path[t] = yieldv
        NW_path[t]    = NW
        Btil_path[t]  = Btilde
        Bhat_path[t]  = Bhat
        EMSP_path[t]  = SPv
        Ct_path[t]    = Ct
        TB_path[t]    = TB*100
        C_Y_path[t]   = C_Y*100
        HYSP_path[t]  = itp_HYSP[omg_path[t], aDMpv]
        Bt_path[t+1]  = Bt1

        ADM_path[t+1] = aDMpv

        Λt1_mkt = Λt1[cm.num_nodes+1:end]
        defmeas_path[t] = 1.0 - sum(Λt1_mkt)

        Λt = copy(Λt1)

    end

    MV_path = simulate_mktval(cm, asm, omg_path, ADM_path, NW_path)
    CA_path = TB_path - (1-cm.ξ)*Bt_path[1:end-1]./exp.(YEM_path)


    plot_YEM_md_stn  = YEM_path/mms_in_mod[1,2]
    plot_omg_md_stn  = log.(omg_path)/cm.σω
    plot_MV_md_stn   = (log.(MV_path)-log.(MV_path_NS)+log.(MV_path_NS[end])-mms_in_mod[2,1]) / mms_in_mod[2,2]
    plot_Ct_md_stn   = (log.(Ct_path)-log.(Ct_path_NS)+log.(Ct_path_NS[end])-mms_in_mod[3,1]) / mms_in_mod[3,2]       

    plot_YEM_dt_stn  = YEM_path/mms_in_mod[1,2]
    plot_MV_dt_stn   = (data_in[:, 3]-mms_in_dat[2,1])/mms_in_dat[2,2]
    plot_Ct_dt_stn   = (data_in[:, 4]-mms_in_dat[3,1])/mms_in_dat[3,2]

    plot_YEM_md_dev  = YEM_path - mms_in_mod[1,1]
    plot_MV_md_dev   = log.(MV_path)-log.(MV_path_NS)+log.(MV_path_NS[end]) - mms_in_mod[2,1]
    plot_Ct_md_dev   = log.(Ct_path)-log.(Ct_path_NS)+log.(Ct_path_NS[end]) - mms_in_mod[3,1]

    plot_YEM_dt_dev  = YEM_path*(mms_in_dat[1,2]/mms_in_mod[1,2]) - mms_in_dat[1,1]
    plot_MV_dt_dev   = data_in[:, 3]-mms_in_dat[2,1]
    plot_Ct_dt_dev   = data_in[:, 4]-mms_in_dat[3,1]

    plot_EMSP = EMSP_path-EMSP_path_NS+EMSP_path_NS[end]
    plot_NW   = log.(NW_path)#-log.(NW_path_NS)+log.(NW_path_NS[1])
    plot_ADM  = ADM_path-ADM_path_NS+ADM_path_NS[end]
    plot_D    = D_path-D_path_NS+D_path_NS[end]
    plot_SPDF = SPDF_path-SPDF_path_NS+SPDF_path_NS[end]

    plot_Bhat = Bhat_path-Bhat_path_NS+Bhat_path_NS[end]
    plot_Btil = Btil_path-Btil_path_NS+Btil_path_NS[end]
    plot_HYSP = HYSP_path-HYSP_path_NS+HYSP_path_NS[end]
    plot_Bt   = Bt_path-Bt_path_NS+Bt_path_NS[end]

    plot_TB   = TB_path-TB_path_NS+TB_path_NS[end]
    plot_C_Y  = C_Y_path-C_Y_path_NS+C_Y_path_NS[end]
    plot_CA   = CA_path-CA_path_NS+CA_path_NS[end]

    plot_yield   = yield_path-yield_path_NS+yield_path_NS[end]
    plot_defmeas = defmeas_path-defmeas_path_NS+defmeas_path_NS[end]


    Nvars = 16
    out_mat = zeros(capT+1, Nvars)

    out_mat[1:capT, 1:7] = [Btil_path Bhat_path plot_yield plot_EMSP HYSP_path plot_NW plot_MV_md_stn]
    out_mat[:, 8]        = ADM_path#plot_ADM
    out_mat[1:capT, 9]   = plot_Ct_md_stn
    out_mat[1:capT, 10]  = plot_TB
    out_mat[1:capT, 11]  = plot_Bt[1:end-1]
    out_mat[1:capT, 12]  = plot_MV_md_dev
    out_mat[1:capT, 13]  = plot_Ct_md_dev
    out_mat[1:capT, 14]  = plot_SPDF
    out_mat[1:capT, 15]  = plot_defmeas
    out_mat[1:capT, 16]  = plot_CA
    

    return out_mat

end


# Amplification Exercise
function compute_amplification_proxy(cm::CoreModule, asm::AshocksModuleOBC, init_val, omg_path, YEM_path, ZEM_path, Λerg, itp_qDFRSK)

    # Create interpolation/extrapolation object
    knots    = (cm.bgrid, cm.zgrid, asm.YEMgrid, cm.ωgrid, cm.aDMgrid)
    itp_def  = interpolate(knots, asm.defmat,     Gridded(Linear()))
    itp_q    = interpolate(knots, asm.q_mat,  Gridded(Linear()))
    itp_bp   = interpolate(knots, asm.bprime_mat, Gridded(Linear()))
    itp_ReDM = interpolate((cm.ωgrid, cm.aDMgrid), asm.ReDM, Gridded(Linear()))
    itp_Red  = interpolate((cm.ωgrid, cm.aDMgrid), asm.Red, Gridded(Linear()))

    capT  = size(omg_path, 1)

    Btil_path, Bhat_path, YLD_path, EMSP_path, NW_path, Ct_path, RDM_path = zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT)

    ADM_path    = zeros(capT+1) # time t
    D_path      = zeros(capT+1) # time t
    ADM_path[1] = init_val[1]
    D_path[1]   = init_val[2]

    Λt = copy(Λerg)
    Bt = dot(Λt[cm.num_nodes+1:end,1], cm.snodes[:,1])
    b_path = zeros(capT+1); b_path[1] = Bt # 0.035 # 0.042 # Bt
    c_path = zeros(capT)
    ReDM_path = zeros(capT)
    Eiotanu_path = zeros(capT)
    q_path = zeros(capT)



    # Bounds for finding ADM
    bounds = [cm.aDMgrid[1]*0.9; asm.ADMmax]


    for t=1:capT
        out_X = solve_system_calib(cm, asm, [YEM_path[t]; omg_path[t]; ADM_path[t]; D_path[t]], Λt, itp_def, itp_q, itp_bp, itp_ReDM, itp_Red, bounds, itp_qDFRSK) # (out_obj, Λt1, defvec_out)
        # out_obj = [aDMp_v; Btilv; Bhatv; Bt1; NW; ReDM; SPv; yield_v; by_avg; sdby; sdSP; Ct; SPDFRK; Dpv; aDMpv2[1]; TB; C_Y]

        aDMpv, Btilde, Bhat, Bt1, NW, ReDM, SPv, yieldv = out_X[1][1], out_X[1][2], out_X[1][3], out_X[1][4], out_X[1][5], out_X[1][6], out_X[1][7], out_X[1][8]
        by_avg, sdby, sdSP, Ct, TB, C_Y = out_X[1][9], out_X[1][10], out_X[1][11], out_X[1][12], out_X[1][16], out_X[1][17]
        D_path[t+1] = out_X[1][14]
  

        Λt1 = out_X[2]

        def_v = itp_def[b_path[t], ZEM_path[t], YEM_path[t], omg_path[t], aDMpv]
        if def_v.<=0.5
            bp_v    = max(itp_bp[b_path[t], ZEM_path[t], YEM_path[t], omg_path[t], aDMpv], cm.bgrid[1])
            qv      = max(0.0, itp_q[bp_v, ZEM_path[t], YEM_path[t], omg_path[t], aDMpv])
            SPv     = (1.0/qv+cm.ξ-1.0) - (1.0/cm.βLD-1.0)
            EMSP_path[t]    = SPv
            Eiotanu_path[t] = qv * itp_ReDM[ZEM_path[t], aDMpv]
            q_path[t] = qv

            c_path[t]   = max(exp(YEM_path[t]+ZEM_path[t]) - b_path[t] + qv*(bp_v-cm.ξ*b_path[t]), 1e-14)
            b_path[t+1] = bp_v
        end


        RDM_path[t]   = fun_RDM(cm.α, cm.δ, ADM_path[t], omg_path[t])

        
        ADM_path[t+1] = aDMpv

        YLD_path[t]  = yieldv
        NW_path[t]   = NW
        Btil_path[t] = Btilde
        Bhat_path[t] = Bhat
        Ct_path[t]   = Ct

        ADM_path[t+1] = aDMpv
        ReDM_path[t] = itp_ReDM[ZEM_path[t], aDMpv]

        Λt = copy(Λt1)

    end

    out_mat = zeros(capT+1, 12)

    out_mat[1:capT, 1:6]    = [Btil_path Bhat_path YLD_path EMSP_path NW_path RDM_path]
    out_mat[:, 7:8]         = [ADM_path b_path]
    out_mat[1:capT,9]       = c_path
    out_mat[1:capT,10]      = ReDM_path
    out_mat[1:capT,11]      = Eiotanu_path
    out_mat[1:capT,12]      = q_path

    return out_mat
end


# Elasticity Exercise
function compute_IRF_phi(cm::CoreModule, asm::AshocksModuleOBC, init_val, omg_path, YEM_path, Λerg, itp_qDFRSK, itp_qIR; secmkt = false)

    # Create interpolation/extrapolation object
    knots    = (cm.bgrid, cm.zgrid, asm.YEMgrid, cm.ωgrid, cm.aDMgrid)
    itp_def  = interpolate(knots, asm.defmat,     Gridded(Linear()))
    itp_q    = interpolate(knots, asm.q_mat,  Gridded(Linear()))
    itp_bp   = interpolate(knots, asm.bprime_mat, Gridded(Linear()))
    itp_ReDM = interpolate((cm.ωgrid, cm.aDMgrid), asm.ReDM, Gridded(Linear()))
    itp_Red  = interpolate((cm.ωgrid, cm.aDMgrid), asm.Red, Gridded(Linear()))

    capT  = size(omg_path, 1)
    Nvars = 10

    Btil_path, Bhat_path, YLD_path, EMSP_path, HYSP_path, NW_path, Ct_path = zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    SPIR_path, SPDFRK_path, Bt1_path = zeros(capT), zeros(capT), zeros(capT)

    ADM_path    = zeros(capT+1) # time t
    D_path      = zeros(capT+1) # time t
    ADM_path[1] = init_val[1]
    D_path[1]   = init_val[2]

    Λt   = copy(Λerg)
    Λtm1 = copy(Λt)


    # Bounds for finding ADM
    bounds = [cm.aDMgrid[1]*0.9; asm.ADMmax]


    for t=1:capT
        out_X = solve_system_calib(cm, asm, [YEM_path[t]; omg_path[t]; ADM_path[t]; D_path[t]], Λt, itp_def, itp_q, itp_bp, itp_ReDM, itp_Red, bounds, itp_qDFRSK)
        # [aDMp_v; Btilv; Bhatv; Bt1; NW; ReDM; SPv; yield_v; by_avg; sdby; sdSP; Ct; SPDFRK; Dpv; aDMp_out; TB; C_Y]
        #   1       2      3      4   5    6     7     8        9      10    11   12   13     14     15      16   17

        aDMpv, Btilde, Bhat, Bt1, NW, ReDM, SPv, yieldv = out_X[1][1], out_X[1][2], out_X[1][3], out_X[1][4], out_X[1][5], out_X[1][6], out_X[1][7], out_X[1][8]
        by_avg, sdby, sdSP, Ct = out_X[1][9], out_X[1][10], out_X[1][11], out_X[1][12]
        D_path[t+1] = out_X[1][14]

        Λt1 = out_X[2]

        SPDFRK_path[t] = out_X[1][13]                                 # spread purely default risk
        qIR_val        = itp_qIR[YEM_path[t], omg_path[t], aDMpv]     # price without default risk
        SPIR_path[t]   = (1.0/qIR_val+cm.ξ-1.0) - (1.0/cm.βLD-1.0)    # spread without default risk

        YLD_path[t]  = yieldv
        NW_path[t]   = NW
        Btil_path[t] = Btilde
        Bhat_path[t] = Bhat
        EMSP_path[t] = SPv
        Ct_path[t]   = Ct
        Bt1_path[t]  = Bt1

        ADM_path[t+1] = aDMpv

        Λtm1 = copy(Λt)
        Λt   = copy(Λt1)

    end

    MV_path = simulate_mktval(cm, asm, omg_path, ADM_path, NW_path)

    ΔMV  = log(MV_path[2])-log(MV_path[1])
    ΔSP  = EMSP_path[2] - EMSP_path[1]
    ΔNW  = log(NW_path[2]) - log(NW_path[1])
    ΔYLD = YLD_path[2] - YLD_path[1]

    ΔSPIR = SPIR_path[2] - SPIR_path[1]
    ΔSPDF = SPDFRK_path[2] - SPDFRK_path[1]

    ΔBt1  = Bt1_path[2] - Bt1_path[1]

    beta = ΔSP / ΔMV

    if secmkt==false
        out = [ΔNW; ΔMV; ΔSP; ΔYLD; beta; ΔSPIR; ΔSPDF; ΔBt1]
    else
        out = ([ADM_path[2]; ADM_path[end]; Btil_path[end]; D_path[end-1]; MV_path[1]], Λtm1)
    end

    return out

end



# Liquidity Policy Exercise
function compute_fall_omega(cm::CoreModule, asm::AshocksModuleOBC, init_val, omg_path, YEM_path, Λerg, itp_qDFRSK)

    # Create interpolation/extrapolation object
    knots       = (cm.bgrid, cm.zgrid, asm.YEMgrid, cm.ωgrid, cm.aDMgrid)
    itp_def     = interpolate(knots, asm.defmat,     Gridded(Linear()))
    itp_q       = interpolate(knots, asm.q_mat,  Gridded(Linear()))
    itp_bp      = interpolate(knots, asm.bprime_mat, Gridded(Linear()))
    itp_ReDM    = interpolate((cm.ωgrid, cm.aDMgrid), asm.ReDM, Gridded(Linear()))
    itp_Red     = interpolate((cm.ωgrid, cm.aDMgrid), asm.Red, Gridded(Linear()))

    capT  = size(omg_path, 1)
    NS    = size(cm.snodes, 1)
    

    ### Simulation in absence of shocks
    Btil_path_NS, Bhat_path_NS, yield_path_NS, EMSP_path_NS, NW_path_NS, Ct_path_NS = zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    TB_path_NS, C_Y_path_NS, SPDF_path_NS, defmeas_path_NS = zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    Bt_path_NS = zeros(capT+1)


    ADM_path_NS    = zeros(capT+1) # time t
    D_path_NS      = zeros(capT+1) # time t
    ADM_path_NS[1] = init_val[1]
    D_path_NS[1]   = init_val[2]

    Λerg_mkt = Λerg[cm.num_nodes+1:end]
    sum_Λ   = sum(Λerg_mkt)

    Bt = 0.0
    for i_s=1:cm.num_nodes
        Bt    += Λerg_mkt[i_s] * cm.snodes[i_s, 1]
    end

    Bt_path_NS[1] = Bt * cm.μEM / sum_Λ

    YEM_path_NS = zeros(YEM_path)
    omg_path_NS = ones(capT)

    Λt = copy(Λerg)


    # Bounds for finding ADM
    bounds = [cm.aDMgrid[1]*0.9; asm.ADMmax]

    for t=1:capT

        state_in = [YEM_path_NS[t]; omg_path_NS[t]; ADM_path_NS[t]; D_path_NS[t]]
        out_X    = solve_system_calib(cm, asm, state_in, Λt, itp_def, itp_q, itp_bp, itp_ReDM, itp_Red, bounds, itp_qDFRSK)
        # out_obj = [aDMp_v; Btilv; Bhatv; Dt1; NW; ReDM; SPv; yield_v; by_avg; sdby; sdSP; Ct; SPDFRK; Dpv; aDMpv2[1]; TB; C_Y]

        aDMpv, Btilde, Bhat, Bt1, NW, ReDM, SPv, yieldv = out_X[1][1], out_X[1][2], out_X[1][3], out_X[1][4], out_X[1][5], out_X[1][6], out_X[1][7], out_X[1][8]
        by_avg, sdby, sdSP, Ct, TB, C_Y = out_X[1][9], out_X[1][10], out_X[1][11], out_X[1][12], out_X[1][16], out_X[1][17]
        D_path_NS[t+1] = out_X[1][14]
        
        Λt1 = out_X[2]

        yield_path_NS[t] = yieldv
        NW_path_NS[t]    = NW
        Btil_path_NS[t]  = Btilde
        Bhat_path_NS[t]  = Bhat
        EMSP_path_NS[t]  = SPv
        Ct_path_NS[t]    = Ct
        TB_path_NS[t]    = TB*100
        C_Y_path_NS[t]   = C_Y*100
        Bt_path_NS[t+1]  = Bt1
        SPDF_path_NS[t]  = out_X[1][13]

        ADM_path_NS[t+1] = aDMpv

        Λt1_mkt = Λt1[cm.num_nodes+1:end]
        defmeas_path_NS[t] = 1.0 - sum(Λt1_mkt)

        Λt = copy(Λt1)

    end


    MV_path_NS = simulate_mktval(cm, asm, omg_path_NS, ADM_path_NS, NW_path_NS)


    ### Simulation given shocks
    Btil_path, Bhat_path, yield_path, EMSP_path, NW_path, Ct_path = zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    TB_path, C_Y_path, SPDF_path, defmeas_path = zeros(capT), zeros(capT), zeros(capT), zeros(capT)
    Bt_path = zeros(capT+1)

    ADM_path    = zeros(capT+1) # time t
    D_path      = zeros(capT+1) # time t
    ADM_path[1] = init_val[1]
    D_path[1]   = init_val[2]

    Bt_path[1]  = Bt_path_NS[1]

    Λt = copy(Λerg)

    for t=1:capT

        state_in = [YEM_path[t]; omg_path[t]; ADM_path[t]; D_path[t]]
        out_X = solve_system_calib(cm, asm, state_in, Λt, itp_def, itp_q, itp_bp, itp_ReDM, itp_Red, bounds, itp_qDFRSK)

        aDMpv, Btilde, Bhat, Bt1, NW, ReDM, SPv, yieldv = out_X[1][1], out_X[1][2], out_X[1][3], out_X[1][4], out_X[1][5], out_X[1][6], out_X[1][7], out_X[1][8]
        by_avg, sdby, sdSP, Ct, TB, C_Y = out_X[1][9], out_X[1][10], out_X[1][11], out_X[1][12], out_X[1][16], out_X[1][17]
        SPDF_path[t] = out_X[1][13]
        D_path[t+1]  = out_X[1][14]

        Λt1 = out_X[2]

        yield_path[t] = yieldv
        NW_path[t]    = NW
        Btil_path[t]  = Btilde
        Bhat_path[t]  = Bhat
        EMSP_path[t]  = SPv
        Ct_path[t]    = Ct
        TB_path[t]    = TB*100
        C_Y_path[t]   = C_Y*100
        Bt_path[t+1]  = Bt1

        ADM_path[t+1] = aDMpv

        Λt1_mkt = Λt1[cm.num_nodes+1:end]
        defmeas_path[t] = 1.0 - sum(Λt1_mkt)

        Λt = copy(Λt1)

    end

    MV_path = simulate_mktval(cm, asm, omg_path, ADM_path, NW_path)


    plot_mv      = log.(MV_path)-log.(MV_path_NS)+log.(MV_path_NS[end])
    plot_Ct      = log.(Ct_path)-log.(Ct_path_NS)+log.(Ct_path_NS[end])
    plot_EMSP    = EMSP_path-EMSP_path_NS+EMSP_path_NS[end]
    plot_ADM     = ADM_path-ADM_path_NS+ADM_path_NS[end]
    plot_D       = D_path-D_path_NS+D_path_NS[end]
    plot_SPDF    = SPDF_path-SPDF_path_NS+SPDF_path_NS[end]

    plot_Bhat    = Bhat_path-Bhat_path_NS+Bhat_path_NS[end]
    plot_Btil    = Btil_path-Btil_path_NS+Btil_path_NS[end]
    plot_Bt      = Bt_path-Bt_path_NS+Bt_path_NS[end]

    plot_TB      = TB_path-TB_path_NS+TB_path_NS[end]
    plot_C_Y     = C_Y_path-C_Y_path_NS+C_Y_path_NS[end]

    plot_yield   = yield_path-yield_path_NS+yield_path_NS[end]
    plot_defmeas = defmeas_path-defmeas_path_NS+defmeas_path_NS[end]


    Nvars = 11
    out_mat = zeros(capT+1, Nvars)

    out_mat[1:capT, 1:5]  = [plot_Btil plot_Bhat plot_yield plot_EMSP plot_mv]
    out_mat[1:capT+1, 6]  = plot_ADM
    out_mat[1:capT, 7:11] = [plot_Ct plot_TB plot_Bt[1:end-1] plot_SPDF plot_defmeas]


    return out_mat

end


