
### Code to define all types to be used ###

using Distributions, Interpolations, Optim, Roots, FastGaussQuadrature, Plots, JLD, DataFrames

type CoreModuleSC
    # Model Parameters
    β::Float64
    α::Float64
    δ::Float64
    γ::Float64
    ρz::Float64
    σz::Float64
    ρω::Float64
    σω::Float64
    ρYEM::Float64
    σYEM::Float64
    ρEMDM::Float64
    θ::Float64
    d0::Float64
    d1::Float64
    κ::Float64
    σn::Float64
    phi::Float64
    Eω::Float64
    μEM::Float64
    μω::Float64
    σM::Float64
    ξ::Float64
    ph0::Float64
    ph1::Float64

    # Grid Parameters
    nz::Int
    zn_mk_fine::Int
    nb::Int
    n_ut::Int
    num_nodes::Int
	naDM::Int
    nω::Int

    bgrid::Vector{Float64}
    zgrid::Vector{Float64}
    zgrid_fine::LinSpace
	aDMgrid::Vector{Float64}
    ωgrid::Vector{Float64}

    Πz::Array{Float64, 2}
    Πω::Array{Float64, 2}
    Πz_fine::Array{Float64, 2}

    zero_ind::Int64

    # Value functions
    vf::Array{Float64, 3}
    vd::Array{Float64, 2}
    vc::Array{Float64, 3}
    bprime_mat::Array{Float64, 3}
    def_states::Array{Float64, 3}
    q_mat::Array{Float64, 3}
    ReDM::Vector{Float64}

    # Quadratures
    quad_x::Array{Float64}
    quad_w::Array{Float64}

    ## For aggregates
    b_grid_fine::Array{Float64}
    snodes::Array{Float64, 2}
    nbar::Float64

    ## Steady State
    Btlstar::Float64
    Bhtstar::Float64
    aDMstar::Float64
    Astar::Float64
    NWstar::Float64
    Dstar::Float64
    REMstar::Float64
    RDMstar::Float64
    nustar::Float64
    Bstar::Float64
    Λstar::SparseVector{Float64, Int64}
    iotas_star::SparseVector{Float64, Int64}

    ## Forecast Rules
    coeffaDM::Vector{Float64}

    maxit::Int64
    damp::Float64
    vftol::Float64
end


type AshocksModuleSC
    # Parameters
    nYEM::Int64
    ρYEM::Float64
    σYEM::Float64
    n_et::Int
    ADMmax::Vector{Float64}

    # Arrays
    YEMgrid::Vector{Float64}
    yt_vec::Vector{Float64}

    # Value functions
    vf::Array{Float64, 5}
    vd::Array{Float64, 4}
    vc::Array{Float64, 5}
    bprime_mat::Array{Float64, 5}
    q_mat::Array{Float64, 5}
    defmat::Array{Float64, 5}
    vs::Array{Float64, 2}      # Pseudo-value function of the lender
    ReDM::Array{Float64, 2}    # Expected excess return for High Yield
    Red::Array{Float64, 2}
    
    # Transition Probabilities
    ΠYEM::Array{Float64, 2}

    # Quadratures
    R::Array{Float64, 2}
    quad_x::Array{Float64}
    quad_w::Array{Float64}

end


function fill_sparse_fun(grid::Vector, pol::Vector)

    # This function takes as input the vector b_{i} and b_prime_{i}
    # and yields the needed values for constructing the sparse matrix.

    #== Initialize vectors ==#
    I_vec   = collect( 1:length(pol) )
    J_vec   = similar(I_vec)

    w_right = Array{Real}(length(pol))


    for (i,value) in enumerate(pol)
        indj   = min(max( searchsortedfirst(grid, value), 2),length(grid))  ## NOTE: 1st value ≥ in grid

        J_vec[i]   = indj
        value      = max( min(value, grid[end]), grid[1])
        w_right[i] = ( value - grid[indj-1] ) / ( grid[indj] - grid[indj-1] )
    end

    #== Repeat indices ==#
    I_vec   = [I_vec; I_vec ]
    J_vec   = [J_vec; J_vec-1 ]
    weights = [w_right; 1.0 - w_right]

    return I_vec, J_vec, weights
end


function cell_to_mat(in_cell)

    NN    = length(in_cell)
    index = zeros(NN)
    Nvars = 0
    capT  = 0

    for i_n = 1:NN
        temp_obj = in_cell[i_n]
        if size(temp_obj, 2) == 1
            capT = max(capT, size(temp_obj, 1))
            index[i_n] = 1
            Nvars += 1
        end
        
    end

    out_mat = zeros(capT, Nvars)

    for i_n=1:Nvars
        if index[i_n] == 1
            vec_vals    = vec(in_cell[i_n])
            length_tmp  = length(vec_vals)
            out_mat[1:length_tmp, i_n] = vec(in_cell[i_n])
        end
    end

    return out_mat
end


sc_RF_fun(cm::CoreModuleSC, ω_v) = max.(min.(cm.ph0.*(ω_v.^(cm.ph1)),1.04),1.005)-1.0;



